import os
from dotenv import load_dotenv
from pydantic import BaseModel, ValidationError
from typing import Optional


def load_and_validate_env(
    Config: BaseModel, file_path: Optional[str] = None
) -> Optional[BaseModel]:
    if file_path:
        load_dotenv(file_path)
    else:
        load_dotenv()

    # Create a env var dictionary
    env_vars = {
        key: os.getenv(key) for key in Config.__annotations__.keys() if os.getenv(key)
    }

    # Validate env vars against ApiConfig model
    try:
        config = Config(**env_vars)
    except ValidationError as e:
        print(f"Invalid configuration: {e}")
        return None

    return config



"""
Show Model Information
Data models for `POST /api/show` endpoint request and response.
Ollama endpoint reference: https://github.com/ollama/ollama/blob/main/docs/api.md#show-model-information
"""

from pydantic import BaseModel
from typing import Optional


class ShowRequest(BaseModel):
    """Request model for showing model information."""

    model: str
    verbose: Optional[bool] = False



import json
import ollama
from oshepherd.worker.app import celery_app
from oshepherd.worker.ollama_task import OllamaCeleryTask
from oshepherd.common.ollama import serialize_ollama_res
from oshepherd.common.redis_service import RedisService
from oshepherd.worker.config import WorkerConfig
from oshepherd.common.lib import load_and_validate_env


@celery_app.task(
    name="oshepherd.worker.tasks.exec_completion",
    bind=True,
    base=OllamaCeleryTask,
)
def exec_completion(self, request_str: str):
    try:
        request = json.loads(request_str)
        print(f"# exec_completion request {request}")
        req_type = request["type"]
        req_payload = request["payload"]
        task_id = self.request.id
        is_streaming = req_payload.get("stream", False)

        # Initialize Redis service for streaming
        redis_service = None
        if is_streaming:
            config = load_and_validate_env(WorkerConfig)
            redis_service = RedisService(config.CELERY_BACKEND_URL)
            stream_channel = f"oshepherd:stream:{task_id}"
            print(f" > Streaming enabled on channel: {stream_channel}")

        if req_type == "generate":
            if is_streaming:
                # Stream responses via Redis Pub/Sub
                for chunk in ollama.generate(**req_payload):
                    serializable_chunk = serialize_ollama_res(chunk)
                    redis_service.publish(
                        stream_channel, json.dumps(serializable_chunk)
                    )
                    # print(f"  > Sending chunk [{task_id}]: done={serializable_chunk.get('done')}")

                # Return success indicator for the celery task
                serializable_response = {"status": "streaming_completed"}
            else:
                response = ollama.generate(**req_payload)
                serializable_response = serialize_ollama_res(response)

        elif req_type == "chat":
            if is_streaming:
                # Stream responses via Redis Pub/Sub
                for chunk in ollama.chat(**req_payload):
                    serializable_chunk = serialize_ollama_res(chunk)
                    redis_service.publish(
                        stream_channel, json.dumps(serializable_chunk)
                    )
                    # print(f"  > Sending chunk [{task_id}]: done={serializable_chunk.get('done')}")

                # Return success indicator for the celery task
                serializable_response = {"status": "streaming_completed"}
            else:
                response = ollama.chat(**req_payload)
                serializable_response = serialize_ollama_res(response)

        elif req_type == "embeddings":
            response = ollama.embeddings(**req_payload)
            serializable_response = serialize_ollama_res(response)

        print(f"  $ success {serializable_response}")
    except Exception as error:
        print(f"  * error exec_completion {error}")
        serializable_response = {
            "error": {"type": str(error.__class__.__name__), "message": str(error)}
        }
        print(
            f"    * error response {serializable_response}",
        )

        if is_streaming and redis_service:
            try:
                error_chunk = {
                    "error": serializable_response["error"]["message"],
                    "done": True,
                }
                redis_service.publish(stream_channel, json.dumps(error_chunk))
            except Exception as publish_error:
                print(f"  * error publishing error to stream: {publish_error}")

        raise

    return serializable_response



from typing import Any, Union


def serialize_ollama_res(response: Any) -> Union[dict, Any]:
    if hasattr(response, "model_dump"):
        return response.model_dump()
    elif hasattr(response, "__dict__"):
        return dict(response)
    else:
        return response



"""
Show Model Information
API implementation of `POST /api/show` endpoint, showing information about a model.
Ollama endpoint reference: https://github.com/ollama/ollama/blob/main/docs/api.md#show-model-information
"""

from fastapi import Request
from oshepherd.api.utils import streamify_json
from oshepherd.api.show.models import ShowRequest
from oshepherd.api.network_data import NetworkData


def load_show_routes(app):
    network_data: NetworkData = app.network_data

    @app.post("/api/show")
    async def show(request: Request):
        request_json = await request.json()
        print(f" # show request: {request_json}")
        show_request = ShowRequest(**request_json)

        ollama_res = network_data.get_model_info(show_request.model)

        status = 200
        if ollama_res.get("error"):
            status = (
                404 if "not found" in ollama_res.get("message", "").lower() else 500
            )

        print(f" $ ollama response [{status}]: {ollama_res}")

        return streamify_json(ollama_res, status)

    return app



"""
List Running Models
API implementation of `GET /api/ps` endpoint, listing models currently loaded in memory across all Oshepherd workers.
Ollama endpoint reference: https://github.com/ollama/ollama/blob/main/docs/api.md#list-running-models
"""

from fastapi import Request
from oshepherd.api.utils import streamify_json
from oshepherd.api.network_data import NetworkData


def load_ps_routes(app):
    network_data: NetworkData = app.network_data

    @app.get("/api/ps")
    async def ps(request: Request):
        print(" # ps request")

        ollama_res = network_data.get_running_models()
        status = 200
        if ollama_res.get("error"):
            ollama_res = {
                "error": "Internal Server Error",
                "message": "error requesting running models",
            }
            status = 500

        print(f" $ ollama response [{status}]: {ollama_res}")

        return streamify_json(ollama_res, status)

    return app



"""
Generate a chat completion
API implementation of `POST /api/chat` endpoint, handling completion orchestration, as replica of the same Ollama server endpoint.
Ollama endpoint reference: https://github.com/ollama/ollama/blob/main/docs/api.md#generate-a-chat-completion
"""

import time
import json
from fastapi import Request
from fastapi.responses import StreamingResponse
from oshepherd.api.utils import streamify_json
from oshepherd.api.chat.models import ChatRequest
from oshepherd.common.redis_service import RedisService


def load_chat_routes(app):

    @app.post("/api/chat")
    async def chat(request: Request):
        from oshepherd.worker.tasks import exec_completion

        request_json = await request.json()
        print(f" # request json: {request_json}")
        chat_request = ChatRequest(**{"payload": request_json})

        # req as json string ready to be sent through broker
        chat_request_json_str = chat_request.model_dump_json()
        print(f" # chat request {chat_request_json_str}")

        is_streaming = request_json.get("stream", False)

        # queue request to remote ollama api server
        task = exec_completion.delay(chat_request_json_str)
        task_id = task.id

        if is_streaming:
            print(f" > Streaming mode enabled for task {task_id}")

            def stream_generator():
                redis_service = RedisService(app.celery_app.conf.broker_url)
                stream_channel = f"oshepherd:stream:{task_id}"
                print(f" > Subscribing to channel: {stream_channel}")

                try:
                    for chunk in redis_service.subscribe_to_channel(stream_channel):
                        # Each chunk is already a dict from Redis
                        chunk_json = json.dumps(chunk) + "\n"
                        # print(f" > Receiving chunk [{task_id}]: {chunk_json.strip()}")
                        # Send worker chunk response to client
                        yield chunk_json.encode("utf-8")

                        # Break after receiving the final chunk
                        if chunk.get("done") is True:
                            print(f" > Stream completed for task {task_id}")
                            break
                except Exception as e:
                    print(f" ! Error streaming response: {e}")
                    error_response = {
                        "error": f"Streaming error: {str(e)}",
                        "done": True,
                    }
                    yield (json.dumps(error_response) + "\n").encode("utf-8")

            return StreamingResponse(
                stream_generator(),
                media_type="application/x-ndjson",
                status_code=200,
            )
        else:
            print(f" > Celery mode enabled for task {task_id}")

            while not task.ready():
                print(" > waiting for response...")
                time.sleep(1)
            ollama_res = task.get(timeout=1)

            status = 200
            if ollama_res.get("error"):
                ollama_res = {
                    "error": "Internal Server Error",
                    "message": f"error executing completion: {ollama_res['error']['message']}",
                }
                status = 500

            print(f" $ ollama response {status}: {ollama_res}")

            return streamify_json(ollama_res, status)

    return app



import threading
from typing import Any, Optional, Dict, Iterator, Generator
from redis import Redis
from redis.exceptions import ConnectionError as RedisConnectionError
import json


class RedisService:
    """
    Resilient Redis service with automatic reconnection.
    """

    def __init__(self, backend_url: str) -> None:
        self.backend_url: str = backend_url
        self.redis_client: Redis = self._create_redis_client()
        self._connection_lock: threading.Lock = threading.Lock()

    def _create_redis_client(self) -> Redis:
        """Create Redis client with optimal settings for reliability."""
        return Redis.from_url(
            self.backend_url,
            socket_keepalive=True,
            retry_on_timeout=True,
            socket_connect_timeout=5,
            socket_timeout=30,
            max_connections=5,
            socket_keepalive_options={},
        )

    def _ensure_connection(self) -> bool:
        """Ensure Redis connection is healthy, reconnect if needed."""
        try:
            self.redis_client.ping()
            return True
        except (
            RedisConnectionError,
            ConnectionResetError,
            TimeoutError,
            BrokenPipeError,
        ) as e:
            print(f" ! Redis connection lost: {e}")
            with self._connection_lock:
                try:
                    self.redis_client = self._create_redis_client()
                    self.redis_client.ping()
                    print(" > Redis connection restored")
                    return True
                except Exception as reconnect_error:
                    print(f" ! Failed to reconnect to Redis: {reconnect_error}")
                    return False

    def _with_retry(self, operation: Any, *args: Any, **kwargs: Any) -> Any:
        """Execute Redis operation with automatic retry on connection failure."""
        if not self._ensure_connection():
            raise RedisConnectionError("Redis connection unavailable")

        try:
            return operation(*args, **kwargs)
        except (
            RedisConnectionError,
            ConnectionResetError,
            TimeoutError,
            BrokenPipeError,
        ) as e:
            print(f" ! Redis operation failed: {e}")
            if self._ensure_connection():
                try:
                    return operation(*args, **kwargs)
                except Exception as retry_error:
                    print(f" ! Operation failed after reconnection: {retry_error}")
                    raise
            else:
                raise RedisConnectionError("Could not recover Redis connection")

    def hset(
        self,
        name: str,
        key: Optional[str] = None,
        value: Optional[Any] = None,
        mapping: Optional[Dict[str, Any]] = None,
    ) -> int:
        return self._with_retry(
            self.redis_client.hset, name, key, value, mapping=mapping
        )

    def hgetall(self, name: str) -> Dict[str, str]:
        return self._with_retry(self.redis_client.hgetall, name)

    def scan_iter(self, **kwargs: Any) -> Iterator[str]:
        return self._with_retry(self.redis_client.scan_iter, **kwargs)

    def ping(self) -> bool:
        return self._with_retry(self.redis_client.ping)

    def publish(self, channel: str, message: str) -> int:
        return self._with_retry(self.redis_client.publish, channel, message)

    def subscribe_to_channel(
        self, channel: str, timeout: Optional[float] = None
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Subscribe to a Redis Pub/Sub channel and yield messages.

        Args:
            channel: The channel name to subscribe to
            timeout: Optional timeout in seconds for listening to messages

        Yields:
            Dict containing the message data
        """
        pubsub = None
        try:
            pubsub = self.redis_client.pubsub()
            pubsub.subscribe(channel)

            for message in pubsub.listen():
                if message["type"] == "subscribe":
                    continue
                elif message["type"] == "message":
                    try:
                        data = json.loads(message["data"])
                        yield data

                        # Check if this is the end marker
                        if data.get("done") is True:
                            break
                    except json.JSONDecodeError:
                        print(f" ! Failed to decode message: {message['data']}")
                        continue
        except Exception as e:
            print(f" ! Error in subscribe_to_channel: {e}")
            raise
        finally:
            if pubsub:
                pubsub.unsubscribe(channel)
                pubsub.close()



import json
import ollama
import socket
import threading
import time
import uuid
import requests
from datetime import datetime, timezone
from oshepherd.worker.config import WorkerConfig
from oshepherd.common.redis_service import RedisService

OSHEPHERD_WORKER_HOSTNAME = socket.gethostname()
OSHEPHERD_WORKER_UUID = uuid.uuid4().hex
OSHEPHERD_WORKER_DATA_PUSH_INTERVAL = 10  # secs
OSHEPHERD_WORKERS_PREFIX_KEY = "oshepherd_worker:"


class WorkerData:

    def __init__(self, config: WorkerConfig):
        self.backend_url = config.CELERY_BACKEND_URL
        self.config = config
        self.ollama_base_url = config.OLLAMA_BASE_URL
        self.redis_service = RedisService(self.backend_url)
        self.hostname = OSHEPHERD_WORKER_HOSTNAME
        self.worker_uuid = OSHEPHERD_WORKER_UUID
        self.worker_id = f"{self.hostname}-{self.worker_uuid}"

    def get_ollama_version(self):
        res = {}
        try:
            req_res = requests.get(f"{self.ollama_base_url}/api/version")
            req_res.raise_for_status()
            res = req_res.json()
        except Exception as error:
            res = {
                "error": {"type": str(error.__class__.__name__), "message": str(error)}
            }
            print(
                f" *** error ollama version fn: {res}",
            )
        return res

    def get_ollama_list(self):
        res = {}
        try:
            ollama_response = ollama.list()
            # Convert ListResponse to dict for JSON serialization
            res = (
                ollama_response.model_dump()
                if hasattr(ollama_response, "model_dump")
                else dict(ollama_response)
            )
        except Exception as error:
            res = {
                "error": {"type": str(error.__class__.__name__), "message": str(error)}
            }
            print(
                f" *** error ollama list fn: {res}",
            )
        return res

    def get_ollama_ps(self):
        res = {}
        try:
            ollama_response = ollama.ps()
            # Convert ProcessResponse to dict for JSON serialization
            res = (
                ollama_response.model_dump()
                if hasattr(ollama_response, "model_dump")
                else dict(ollama_response)
            )
        except Exception as error:
            res = {
                "error": {"type": str(error.__class__.__name__), "message": str(error)}
            }
            print(
                f" *** error ollama ps fn: {res}",
            )
        return res

    def get_ollama_show(self, model_name):
        """Get show information for a specific model."""
        res = {}
        try:
            ollama_response = ollama.show(model_name)
            # Convert ShowResponse to dict for JSON serialization
            res = (
                ollama_response.model_dump()
                if hasattr(ollama_response, "model_dump")
                else dict(ollama_response)
            )
        except Exception as error:
            res = {
                "error": {"type": str(error.__class__.__name__), "message": str(error)}
            }
            print(
                f" *** error ollama show fn for model {model_name}: {res}",
            )
        return res

    def get_ollama_show_map(self):
        """Get show information for all available models."""
        show_map = {}
        try:
            list_res = self.get_ollama_list()
            models = list_res.get("models", [])

            for model in models:
                model_name = model.get("model")
                if model_name:
                    show_res = self.get_ollama_show(model_name)
                    show_map[model_name] = show_res

            print(
                f" >>> worker {self.worker_id} fetched show data for {len(show_map)} models"
            )
        except Exception as e:
            print(f" ! Failed to fetch show data: {e}")

        return show_map

    def get_data(self):
        version_res = self.get_ollama_version()
        serialized_version_res = json.dumps(version_res, default=str)
        tags_list_res = self.get_ollama_list()
        serialized_tags_list_res = json.dumps(tags_list_res, default=str)
        ps_res = self.get_ollama_ps()
        serialized_ps_res = json.dumps(ps_res, default=str)
        show_map = self.get_ollama_show_map()
        serialized_show_map = json.dumps(show_map, default=str)
        now = datetime.now(timezone.utc).isoformat()

        return {
            "worker_id": self.worker_id,
            "hostname": self.hostname,
            "uuid": self.worker_uuid,
            "version": serialized_version_res,
            "tags": serialized_tags_list_res,
            "ps": serialized_ps_res,
            "show": serialized_show_map,
            "heartbeat": now,
        }

    def push_data(self):
        try:
            worker_data = self.get_data()
            self.redis_service.hset(
                f"{OSHEPHERD_WORKERS_PREFIX_KEY}{self.worker_id}", mapping=worker_data
            )
            print(f" >>> worker {self.worker_id} data pushed")
        except Exception as e:
            print(f" ! Data push failed: {e}")

    def start_data_push(self):
        print(f" > worker {self.worker_id} data push setup starting")

        def run_periodically():
            while True:
                self.push_data()
                time.sleep(OSHEPHERD_WORKER_DATA_PUSH_INTERVAL)

        thread = threading.Thread(target=run_periodically)
        thread.daemon = True
        thread.start()

        print(f" > worker {self.worker_id} data push setup finished")
